import pkg_resources

from amino import Either, Boolean, Path, Maybe, L, _, do, IO, Do
from amino.boolean import false
from amino.case import Case

from chromatin.model.venv import (VenvStatus, VenvExistent, VenvAbsent, cons_venv, VenvPackageStatus,
                                  VenvPackageExistent, VenvPackageAbsent, Venv)
from chromatin.model.rplugin import (DirRplugin, SiteRplugin, VenvRplugin, Rplugin, RpluginStatus, RpluginReady,
                                     RpluginAbsent)


@do(IO[VenvStatus])
def venv_existent(dir: Path, plugin: Rplugin) -> Do:
    venv = yield cons_venv(dir, plugin)
    return VenvExistent(plugin, venv)


@do(IO[VenvStatus])
def check_venv(base_dir: Path, plugin: Rplugin) -> Do:
    dir = base_dir / plugin.name
    exists = yield IO.delay(dir.exists)
    yield (
        venv_existent(dir, plugin)
        if exists else
        IO.pure(VenvAbsent(plugin))
    )


@do(IO[Boolean])
def venv_exists(base_dir: Path, plugin: Rplugin) -> Do:
    status = yield check_venv(base_dir, plugin)
    return isinstance(status, VenvExistent)


@do(IO[VenvPackageStatus])
def venv_package_status(venv: Venv, req: str) -> Do:
    ws = yield IO.delay(pkg_resources.WorkingSet, [venv.site])
    req = yield IO.delay(pkg_resources.Requirement, req)
    return Maybe.check(ws.by_key.get(req.key)) / L(VenvPackageExistent)(venv, _) | VenvPackageAbsent(venv)


def venv_package_status_main(venv: Venv) -> IO[VenvPackageStatus]:
    return venv_package_status(venv, venv.name)


@do(IO[Boolean])
def venv_package_installed(venv: Venv) -> Do:
    status = yield venv_package_status_main(venv)
    return status.exists


class venv_status_check(Case[VenvExistent, IO[Boolean]], alg=VenvStatus):

    def venv_existent(self, status: VenvExistent) -> IO[Boolean]:
        return venv_package_installed(status.venv)

    def venv_absent(self, status: VenvAbsent) -> IO[Boolean]:
        return IO.pure(false)


class rplugin_status(Case[Rplugin, IO[RpluginStatus]], alg=Rplugin):

    def __init__(self, base_dir: Path) -> None:
        self.base_dir = base_dir

    @do(IO[RpluginStatus])
    def dir_rplugin(self, rplugin: DirRplugin) -> Do:
        is_dir = yield IO.delay(Path(rplugin.spec).is_dir)
        return RpluginReady(rplugin) if is_dir else RpluginAbsent(rplugin)

    def site_rplugin(self, rplugin: SiteRplugin) -> IO[RpluginStatus]:
        return IO.pure(
            RpluginReady(rplugin)
            if Either.import_module(rplugin.spec).present
            else RpluginAbsent(rplugin)
        )

    @do(IO[RpluginStatus])
    def venv_rplugin(self, rplugin: VenvRplugin) -> Do:
        venv_status = yield check_venv(self.base_dir, rplugin)
        exists = yield venv_status_check.match(venv_status)
        ctor = exists.cata(RpluginReady, RpluginAbsent)
        return ctor(rplugin)


@do(IO[Boolean])
def rplugin_ready(base_dir: Path, plugin: Rplugin) -> Do:
    status = yield rplugin_status(base_dir)(plugin)
    return isinstance(status, RpluginReady)


__all__ = ('check_venv', 'venv_exists', 'venv_package_installed', 'venv_status_check', 'rplugin_status',
           'rplugin_ready',)



from amino import do, List, _, Lists, L, Path
from amino.do import Do
from amino.boolean import false, true
from amino.io import IOException
from amino.lenses.lens import lens

from ribosome.nvim.io.state import NS
from ribosome.logging import ribo_log
from ribosome.process import SubprocessResult
from ribosome.compute.api import prog
from ribosome.compute.output import Echo
from ribosome.compute.prog import Prog
from ribosome.compute.ribosome_api import Ribo

from chromatin.components.core.logic import (install_plugins, add_installed, reboot_plugins, activate_by_names,
                                             deactivate_by_names, missing_plugins, venv_from_meta)
from chromatin.util import resources
from chromatin.model.venv import Venv, VenvMeta
from chromatin.config.resources import ChromatinResources
from chromatin.env import Env
from chromatin.settings import venv_dir, autoreboot
from chromatin.components.core.trans.tpe import CrmRibosome


@prog.echo
@do(NS[Env, Echo])
def install_result(installed: List[SubprocessResult[Venv]], errors: List[IOException]) -> Do:
    errors.foreach(lambda e: ribo_log.caught_exception('installing plugin', e))
    success, failed = installed.split(_.success)
    success_venvs = success / _.data
    failed_venvs = failed / _.data
    rplugins = success_venvs / _.rplugin
    yield rplugins.traverse(add_installed, NS)
    return (
        Echo.error(resources.plugins_install_failed(failed_venvs / _.name))
        if not failed.empty else
        Echo.unit
        if success.empty else
        Echo.info(resources.installed_plugins(success_venvs / _.name))
    )


@prog.do(None)
def install_missing() -> Do:
    missing = yield missing_plugins()
    yield install_plugins(missing, false)


@prog.echo
@do(NS[ChromatinResources, Echo])
def activate(*plugins: str) -> Do:
    already_installed = yield activate_by_names(Lists.wrap(plugins))
    names = already_installed / _.name
    return (
        Echo.unit
        if already_installed.empty else
        Echo.info(resources.already_active(names))
    )


@prog
def deactivate(*plugins: str) -> NS[Env, None]:
    return deactivate_by_names(Lists.wrap(plugins))


@do(NS[Env, None])
def updated(result: List[Venv]) -> Do:
    venvs = result.filter_not(_.name == 'chromatin')
    reboot = yield NS.lift(autoreboot.value_or_default())
    if reboot:
        yield reboot_plugins(venvs / _.name)


def filter_venvs_by_name(venvs: List[Venv], names: List[str]) -> List[Venv]:
    return venvs if names.empty else venvs.filter(lambda v: v.name in names)


def installed_with_crm() -> NS[Env, List[VenvMeta]]:
    return NS.inspect(lambda a: a.ready.flat_map(a.venvs.lift).cons_m(a.chromatin_venv))


@do(NS[Env, List[Venv]])
def updateable(venv_dir: Path, rplugins: List[str]) -> Do:
    all = yield installed_with_crm()
    selected = all if rplugins.empty else filter_venvs_by_name(all, rplugins)
    yield selected.traverse(L(venv_from_meta)(venv_dir, _), NS)


@prog
@do(NS[CrmRibosome, List[Venv]])
def updateable_venvs(plugins: List[str]) -> Do:
    dir = yield Ribo.setting(venv_dir)
    yield Ribo.zoom_main(updateable(dir, plugins))


@prog.do(None)
def update_plugins_io(plugins: List[str]) -> Do:
    venvs = yield updateable_venvs(plugins)
    yield (
        Prog.error(resources.no_plugins_match_for_update(plugins))
        if venvs.empty else
        install_plugins(venvs, true)
    )


@prog.echo
@do(NS[ChromatinResources, Echo])
def updated_plugins(results: List[SubprocessResult[Venv]], errors: List[IOException]) -> Do:
    errors.foreach(lambda e: ribo_log.caught_exception('updating plugin', e))
    success, failed = results.split(_.success)
    success_venvs = success / _.data
    failed_venvs = failed / _.data
    yield updated(success_venvs)
    return (
        Echo.info(resources.updated_plugins(success_venvs / _.name))
        if failed.empty else
        Echo.error(resources.plugins_install_failed(failed_venvs / _.name))
    )


@prog.do(None)
def update_plugins(*ps: str) -> Do:
    plugins = Lists.wrap(ps)
    success, fail = yield update_plugins_io(plugins)
    yield updated_plugins(success, fail)


@prog
def reboot(*plugins: str) -> NS[ChromatinResources, None]:
    return reboot_plugins(Lists.wrap(plugins)).replace(None)


__all__ = ('install_result', 'install_missing', 'update_plugins')



import sys
from typing import Tuple

from amino import do, __, _, Just, List, Either, Nil, Boolean, Right, Path, Lists, L, IO, Nothing
from amino.do import Do
from amino.state import State
from amino.io import IOException
from amino.func import ReplaceVal
from amino.util.string import camelcaseify, camelcase
from amino.lenses.lens import lens
from amino.case import Case
from amino.logging import module_log

from ribosome.nvim.io.state import NS
from ribosome.process import SubprocessResult, Subprocess
from ribosome.nvim.io.compute import NvimIO
from ribosome.compute.api import prog
from ribosome.compute.output import GatherSubprocesses
from ribosome.nvim.io.api import N
from ribosome.nvim.api.command import runtime, nvim_command, nvim_command_output, nvim_sync_command
from ribosome.nvim.api.exists import wait_for_command, command_exists, wait_until_function_produces
from ribosome.nvim.api.function import nvim_call_json
from ribosome.rpc.define import ActiveRpcTrigger, undef_command
from ribosome.compute.ribosome_api import Ribo
from ribosome.components.internal.prog import RpcTrigger
from ribosome.nvim.api.variable import var_becomes

from chromatin.model.venv import Venv, cons_venv, VenvMeta, cons_venv_under
from chromatin.host import start_host, stop_host
from chromatin.util import resources
from chromatin.model.rplugin import Rplugin, ActiveRplugin, DirRplugin, VenvRplugin, SiteRplugin, ActiveRpluginMeta
from chromatin.config.resources import ChromatinResources
from chromatin.rplugin import venv_package_installed
from chromatin.env import Env
from chromatin.components.core.trans.tpe import CrmRibosome
from chromatin.settings import handle_crm, venv_dir, rplugins, debug_pythonpath

log = module_log()


@do(NvimIO[List[ActiveRpcTrigger]])
def define_handlers(active_rplugin: ActiveRplugin) -> Do:
    channel = active_rplugin.channel
    cname = camelcaseify(active_rplugin.name)
    rpc_handlers_fun = f'{cname}RpcHandlers'
    result = yield N.call_once_defined(rpc_handlers_fun, timeout=3)
    handlers = (
        Lists.wrap(result)
        .flat_map(RpcHandlerSpec.decode)
        .map(lambda spec: ActiveRpcTrigger(spec=spec, channel=channel))
    )
    yield N.pure(handlers)


# FIXME handlers cannot be obtained while the plugin is initializing, as the RpcHandlers function will be reported as
# 'in use' when it is being polled during definition
# fetch handlers when using them, i.e. during shutdown
@do(NS[Env, None])
def activated(active_rplugin: ActiveRplugin) -> Do:
    log.debug(f'activated {active_rplugin}')
    spec = active_rplugin.rplugin
    yield NS.lift(runtime(f'chromatin/{spec.name}/*'))
    yield NS.modify(__.host_started(active_rplugin.meta))


@do(NvimIO[ActiveRplugin])
def start_rplugin_host(rplugin: Rplugin, python_exe: Path, bin_path: Path, plugin_path: Path) -> Do:
    debug_global = yield debug_pythonpath.value
    debug = debug_global.get_or_strict(rplugin.debug)
    channel, pid = yield start_host(python_exe, bin_path, plugin_path, debug, rplugin.pythonpath)
    return ActiveRplugin(rplugin, ActiveRpluginMeta(rplugin.name, channel, pid))


@do(NvimIO[Venv])
def venv_from_rplugin(rplugin: VenvRplugin) -> Do:
    dir = yield venv_dir.value_or_default()
    yield N.from_io(cons_venv_under(dir, rplugin))


class activate_rplugin_io(Case[Rplugin, NvimIO[ActiveRplugin]], alg=Rplugin):

    def dir_rplugin(self, rplugin: DirRplugin) -> NvimIO[ActiveRplugin]:
        python_exe = Path(sys.executable)
        bin_path = python_exe.parent
        plugin_path = Path(rplugin.spec) / '__init__.py'
        return start_rplugin_host(rplugin, python_exe, bin_path, plugin_path)

    @do(NvimIO[ActiveRplugin])
    def venv_rplugin(self, rplugin: VenvRplugin) -> Do:
        venv = yield venv_from_rplugin(rplugin)
        python_exe = yield N.from_either(venv.meta.python_executable)
        bin_path = yield N.from_either(venv.meta.bin_path)
        yield start_rplugin_host(venv.rplugin, python_exe, bin_path, venv.plugin_path)

    # TODO start with module
    def site_rplugin(self, rplugin: SiteRplugin) -> NvimIO[ActiveRplugin]:
        return N.error('site rplugins not implemented yet')


@do(NS[ChromatinResources, None])
def activate_rplugin(rplugin: Rplugin) -> Do:
    active_rplugin = yield NS.lift(activate_rplugin_io.match(rplugin))
    yield activated(active_rplugin).zoom(lens.data)


@do(NS[ChromatinResources, List[Rplugin]])
def activate_multi(new_rplugins: List[Rplugin]) -> Do:
    active = yield NS.inspect_either(_.active_rplugins).zoom(lens.data)
    active_rplugins = active / _.rplugin
    already_active, inactive = new_rplugins.split(active_rplugins.contains)
    yield inactive.traverse(activate_rplugin, NS).replace(already_active)


@do(NS[ChromatinResources, List[Rplugin]])
def activate_by_names(plugins: List[str]) -> Do:
    getter = _.ready_rplugins if plugins.empty else __.ready_by_name(plugins)
    rplugins = yield NS.inspect_either(getter).zoom(lens.data)
    yield (
        NS.error(resources.no_plugins_match_for_activation(plugins))
        if rplugins.empty else
        activate_multi(rplugins)
    )


def activate_all() -> NS[ChromatinResources, List[Rplugin]]:
    return activate_by_names(Nil)


@do(NvimIO[None])
def undef_trigger(trigger: RpcTrigger) -> Do:
    method = undef_command.match(trigger.method)
    yield method.map(lambda a: nvim_command(a, trigger.name)).get_or_strict(N.unit)


@do(NvimIO[None])
def stop_rplugin(name: str, channel: int, triggers: List[RpcTrigger]) -> Do:
    yield nvim_command(f'{camelcase(name)}Quit')
    yield triggers.traverse(undef_trigger, NvimIO)
    yield nvim_command('autocmd!', name)
    yield stop_host(channel)


@do(NS[Env, None])
def deactivate_rplugin(active_rplugin: ActiveRplugin) -> Do:
    rplugin = active_rplugin.rplugin
    meta = active_rplugin.meta
    cname = camelcaseify(rplugin.name)
    rpc_triggers_fun = f'{cname}RpcTriggers'
    triggers = yield NS.lift(nvim_call_json(rpc_triggers_fun))
    yield NS.modify(__.deactivate_rplugin(meta))
    yield NS.lift(stop_rplugin(rplugin.name, meta.channel, triggers))
    log.debug(f'deactivated {active_rplugin}')
    yield NS.pure(None)


def deactivate_multi(rplugins: List[ActiveRplugin]) -> NS[Env, None]:
    return rplugins.traverse(deactivate_rplugin, NS).replace(None)


@do(NS[Env, None])
def deactivate_by_names(plugins: List[str]) -> Do:
    getter = _.active_rplugins if plugins.empty else __.active_rplugins_by_name(plugins)
    rplugins = yield NS.inspect_either(getter)
    yield (
        NS.error(resources.no_plugins_match_for_deactivation(plugins))
        if rplugins.empty else
        deactivate_multi(rplugins)
    )


@do(NS[ChromatinResources, List[Rplugin]])
def reboot_plugins(plugins: List[str]) -> Do:
    yield deactivate_by_names(plugins).zoom(lens.data)
    yield activate_by_names(plugins)


# TODO don't fail if one plugin doesn't start
@do(NS[Env, None])
def activation_complete() -> Do:
    rplugins = yield NS.inspect_either(_.uninitialized_rplugins)
    prefixes = rplugins / _.name / camelcase
    def wait() -> NvimIO[None]:
        return prefixes.traverse(lambda p: wait_until_function_produces(True, f'{p}Poll'), NvimIO)
    @do(NvimIO[None])
    def rplugin_stage(prefix: str, num: int) -> Do:
        cmd = f'{prefix}Stage{num}'
        exists = yield command_exists(cmd)
        if exists:
            yield nvim_command(cmd)
    def stage(num: int) -> NvimIO[None]:
        return prefixes.traverse(L(rplugin_stage)(_, num), NvimIO)
    yield NS.lift(wait())
    yield NS.lift(Lists.range(1, 5).traverse(stage, NvimIO))
    yield NS.modify(__.initialization_complete())


@do(NS[ChromatinResources, None])
def activate_newly_installed() -> Do:
    new = yield NS.inspect_either(_.inactive).zoom(lens.data)
    yield activate_multi(new)
    yield activation_complete().zoom(lens.data)


def add_venv(venv: VenvMeta) -> State[Env, None]:
    return State.modify(__.add_venv(venv))


@do(State[Env, None])
def add_installed(rplugin: Rplugin) -> Do:
    yield State.modify(__.add_installed(rplugin.name))


@do(NS[Env, None])
def already_installed(venv_dir: Path, rplugin: Rplugin) -> Do:
    yield add_installed(rplugin).nvim
    if isinstance(rplugin, VenvRplugin):
        venv = yield NS.from_io(cons_venv(venv_dir, rplugin))
        yield add_venv(venv.meta).nvim


@do(Either[str, Subprocess[Venv]])
def install_venv(venv: Venv) -> Do:
    log.debug(f'installing {venv}')
    bin_path = yield venv.meta.bin_path
    pip_bin = bin_path / 'pip'
    args = List('install', '-U', '--no-cache', venv.req)
    yield Right(Subprocess(pip_bin, args, venv, timeout=60))


@prog.subproc.gather
@do(NS[Env, GatherSubprocesses[Venv]])
def install_plugins_procs(venvs: List[Venv], update: Boolean) -> Do:
    action = 'updating' if update else 'installing'
    log.debug(f'{action} venvs: {venvs}')
    procs = yield NS.from_either(venvs.traverse(install_venv, Either))
    return GatherSubprocesses(procs, timeout=60)


@prog
def install_plugins_result(results: List[Either[IOException, SubprocessResult[Venv]]]
                           ) -> NS[Env, Tuple[List[Venv], List[str]]]:
    venvs = results.flat_map(__.cata(ReplaceVal(Nil), List))
    errors = results.flat_map(__.cata(List, ReplaceVal(Nil)))
    log.debug(f'installed plugins: {venvs}')
    if errors:
        log.debug(f'error installing plugins: {errors}')
    return NS.pure((venvs, errors))


@prog.do(None)
def install_plugins(venvs: List[Venv], update: Boolean) -> Do:
    result = yield install_plugins_procs(venvs, update)
    yield install_plugins_result(result)


@do(NS[CrmRibosome, None])
def add_crm_venv() -> Do:
    handle = yield Ribo.setting(handle_crm)
    if handle:
        log.debug('adding chromatin venv')
        plugin = Rplugin.simple('chromatin')
        yield Ribo.modify_main(__.set.chromatin_rplugin(Just(plugin)))
        dir = yield Ribo.setting(venv_dir)
        venv = yield NS.from_io(cons_venv(dir, plugin))
        yield Ribo.modify_main(__.set.chromatin_venv(Just(venv)))


@do(NS[Env, List[Rplugin]])
def read_conf() -> Do:
    plugin_conf = yield Ribo.setting(rplugins)
    specs = plugin_conf.traverse(Rplugin.from_config, Either)
    yield NS.from_either(specs)


@do(NS[Env, List[Rplugin]])
def rplugins_with_crm() -> Do:
    rplugins = yield NS.inspect(_.rplugins)
    crm = yield NS.inspect(_.chromatin_rplugin)
    return rplugins.cons_m(crm)


@do(NS[Env, Either[str, Rplugin]])
def rplugin_by_name(name: str) -> Do:
    rplugins = yield rplugins_with_crm()
    return rplugins.find(lambda a: a.name == name).to_either(f'no rplugin with name `{name}`')


@do(NS[Env, Venv])
def venv_from_meta(venv_dir: Path, meta: VenvMeta) -> Do:
    rplugin_e = yield rplugin_by_name(meta.name)
    rplugin = yield NS.from_either(rplugin_e)
    yield NS.from_io(cons_venv_under(venv_dir, rplugin))


@prog
@do(NS[CrmRibosome, List[Venv]])
def missing_plugins() -> Do:
    dir = yield Ribo.setting(venv_dir)
    venv_metas = yield Ribo.inspect_main(_.venvs.v)
    venvs = yield Ribo.zoom_main(venv_metas.traverse(L(venv_from_meta)(dir, _), NS))
    package_status = yield NS.from_io(venvs.traverse(venv_package_installed, IO))
    return venvs.zip(package_status).collect(lambda a: Nothing if a[1] else Just(a[0]))


__all__ = ('add_crm_venv', 'read_conf', 'install_plugins', 'add_installed', 'missing_plugins')



from typing import Type, Any, Tuple
import json

from amino import Either, Map, Regex, do, Do, Try, List, Maybe, Nil, Nothing
from amino.dat import ADT, Dat
from amino.json.decoder import decode_json_type


class ConfigRplugin(Dat['ConfigRplugin']):

    def __init__(self, spec: str, name: Maybe[str], debug: Maybe[bool], pythonpath: Maybe[List[str]]) -> None:
        self.spec = spec
        self.name = name
        self.debug = debug
        self.pythonpath = pythonpath


class Rplugin(ADT['Rplugin']):

    @classmethod
    def cons(
            cls,
            name: str,
            spec: str,
            debug: bool=False,
            pythonpath: List[str]=Nil,
            interpreter: Maybe[str]=Nothing,
    ) -> 'Rplugin':
        return cls(name, spec, debug, pythonpath, interpreter)

    def __init__(
            self,
            name: str,
            spec: str,
            debug: bool,
            pythonpath: List[str],
            interpreter: Maybe[str],
    ) -> None:
        self.name = name
        self.spec = spec
        self.debug = debug
        self.pythonpath = pythonpath
        self.interpreter = interpreter

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and self.name == other.name and self.spec == other.spec

    @staticmethod
    @do(Either[str, 'Rplugin'])
    def from_config(data: dict) -> Do:
        json_conf = yield Try(json.dumps, data)
        conf = yield decode_json_type(json_conf, ConfigRplugin)
        return cons_rplugin(conf)

    @staticmethod
    def simple(name: str) -> 'Rplugin':
        return VenvRplugin.cons(name=name, spec=name)


class VenvRplugin(Rplugin):
    pass


class DirRplugin(Rplugin):
    pass


class SiteRplugin(Rplugin):
    pass


ctors: Map[str, Type[Rplugin]] = Map(
    dir=DirRplugin,
    site=SiteRplugin,
    venv=VenvRplugin,
)
prefixes = ctors.k.mk_string('|')
spec_rex = Regex(f'(?P<prefix>{prefixes}):(?P<spec>.*)')


@do(Either[str, Tuple[Type[Rplugin], str]])
def parse_spec(raw_spec: str) -> Do:
    match = yield spec_rex.match(raw_spec)
    prefix, spec = yield match.all_groups('prefix', 'spec')
    tpe = yield ctors.lift(prefix).to_either('invalid rplugin spec prefix `{prefix}`')
    return tpe, spec


def cons_rplugin(conf: ConfigRplugin) -> Rplugin:
    tpe, spec = parse_spec(conf.spec).get_or_strict((VenvRplugin, conf.spec))
    return tpe.cons(
        conf.name.get_or_strict(spec),
        spec,
        conf.debug.get_or_strict(False),
        conf.pythonpath.get_or_strict(Nil),
    )


class RpluginStatus(ADT['RpluginStatus']):

    def __init__(self, rplugin: Rplugin) -> None:
        self.rplugin = rplugin


class RpluginReady(RpluginStatus):
    pass


class RpluginAbsent(RpluginStatus):
    pass


class ActiveRpluginMeta(Dat['ActiveRpluginMeta']):

    def __init__(self, rplugin: str, channel: int, pid: int) -> None:
        self.rplugin = rplugin
        self.channel = channel
        self.pid = pid


class ActiveRplugin(Dat['ActiveRplugin']):

    def __init__(self, rplugin: Rplugin, meta: ActiveRpluginMeta) -> None:
        self.rplugin = rplugin
        self.meta = meta

    @property
    def name(self) -> str:
        return self.rplugin.name


__all__ = ('Rplugin', 'VenvRplugin', 'DirRplugin', 'SiteRplugin', 'cons_rplugin', 'RpluginStatus', 'RpluginReady',
           'RpluginAbsent', 'ActiveRpluginMeta', 'ActiveRplugin')



import sys
import shutil
from typing import Tuple

from amino import do, __, List, _, Boolean, L, Either, Nil, Right, Left, IO, Maybe, Nothing, Path, env, Lists
from amino.do import Do
from amino.state import State, EitherState
from amino.lenses.lens import lens
from amino.logging import module_log
from ribosome.nvim.io.state import NS
from ribosome import ribo_log
from ribosome.compute.api import prog
from ribosome.compute.output import Echo, GatherIOs
from ribosome.compute.ribosome_api import Ribo
from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.io.api import N

from chromatin.components.core.logic import (add_crm_venv, read_conf, activate_newly_installed, already_installed,
                                             add_venv)
from chromatin.model.rplugin import Rplugin, cons_rplugin, ConfigRplugin
from chromatin.components.core.trans.install import install_missing, install_result
from chromatin.model.rplugin import RpluginReady, VenvRplugin
from chromatin.model.venv import bootstrap, Venv
from chromatin.rplugin import rplugin_status
from chromatin.util import resources
from chromatin.settings import venv_dir, autostart, interpreter
from chromatin.config.resources import ChromatinResources
from chromatin.env import Env
from chromatin.components.core.trans.tpe import CrmRibosome

log = module_log()


@prog
@do(NS[CrmRibosome, List[Rplugin]])
def initialize() -> Do:
    yield add_crm_venv()
    yield Ribo.zoom_main(read_conf())


@prog
@do(EitherState[Env, None])
def bootstrap_result(result: List[Either[str, Venv]]) -> Do:
    def split(z: Tuple[List[str], List[Venv]], a: Either[str, Venv]) -> Tuple[List[str], List[Venv]]:
        err, vs = z
        return a.cata((lambda e: (err.cat(e), vs)), (lambda v: (err, vs.cat(v))))
    errors, venvs = result.fold_left((Nil, Nil))(split)
    ribo_log.debug(f'bootstrapped venvs: {venvs}')
    yield venvs.map(_.meta).traverse(add_venv, State).to(EitherState)
    error = errors.map(str).join_comma
    ret = Left(f'failed to setup venvs: {error}') if errors else Right(None)
    yield EitherState.lift(ret)


@do(NvimIO[Path])
def virtualenv_interpreter(venv: str, executable: str) -> Do:
    path_s = yield N.from_either(env.get('PATH'))
    path = Lists.split(path_s, ':')
    clean_path = path.filter_not(lambda a: a.startswith(venv))
    candidate = yield N.delay(lambda a: shutil.which(executable, path=clean_path.mk_string(':')))
    return Path(candidate)


@do(NvimIO[Path])
def find_interpreter(spec: str) -> Do:
    path = Path(spec)
    exists = yield N.delay(lambda v: path.exists())
    venv = env.get('VIRTUAL_ENV').get_or_strict('[no venv]')
    yield (
        N.pure(path)
        if exists else
        virtualenv_interpreter(venv, spec)
    )


@do(NS[CrmRibosome, Path])
def python_interpreter() -> Do:
    spec = yield Ribo.setting_raw(interpreter)
    yield NS.lift(find_interpreter(spec.get_or_strict('python3.7')))


@prog.io.gather
@do(NS[CrmRibosome, GatherIOs])
def setup_venvs_ios() -> Do:
    dir = yield Ribo.setting(venv_dir)
    plugins = yield Ribo.inspect_main(_.rplugins)
    status = yield NS.from_io(plugins.traverse(rplugin_status(dir), IO))
    ready, absent = status.split_type(RpluginReady)
    yield Ribo.zoom_main((ready / _.rplugin).traverse(L(already_installed)(dir, _), NS))
    absent_venvs, other = (absent / _.rplugin).split_type(VenvRplugin)
    interpreter = yield python_interpreter()
    yield NS.pure(GatherIOs(absent_venvs.map(L(bootstrap)(interpreter, dir, _)), timeout=30))


@prog.do(None)
def setup_venvs() -> Do:
    '''check whether a venv exists for each plugin in the env.
    for those that don't have one, create venvs in `g:chromatin_venv_dir`.
    '''
    result = yield setup_venvs_ios()
    yield bootstrap_result(result)


@prog
def post_setup() -> NS[ChromatinResources, None]:
    return activate_newly_installed()


@prog.do(None)
def setup_plugins() -> Do:
    yield setup_venvs()
    installed, errors = yield install_missing()
    yield install_result(installed, errors)
    yield post_setup()


@prog
@do(NS[ChromatinResources, Boolean])
def add_plugins(plugins: List[Rplugin]) -> None:
    yield NS.modify(__.add_plugins(plugins)).zoom(lens.data)
    yield Ribo.setting(autostart)


@prog.do(None)
def plugins_added(plugins: List[Rplugin]) -> Do:
    autostart = yield add_plugins(plugins)
    if autostart:
        yield setup_plugins()


@prog.do(None)
def init() -> Do:
    plugins = yield initialize()
    yield plugins_added(plugins)


@prog.echo
@do(NS[CrmRibosome, Echo])
def show_plugins() -> Do:
    dir = yield Ribo.setting(venv_dir)
    rplugins = yield Ribo.inspect_main(lambda a: a.rplugins)
    return Echo.info(resources.show_plugins(dir, rplugins))


@prog.do(None)
def add_plugin(spec: str, name) -> Do:
    plugin = cons_rplugin(ConfigRplugin(spec, Maybe.optional(name), Nothing, Nothing))
    yield plugins_added(List(plugin))


__all__ = ('init', 'show_plugins')



from typing import Any

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.io.api import N
from ribosome.rpc.define import ActiveRpcTrigger

from amino import List, _, Either, Boolean, Map, __, Maybe, Nil, L, do, Do
from amino.dat import Dat

from chromatin.model.venv import Venv, cons_venv_under, VenvMeta
from chromatin.model.rplugin import Rplugin, ActiveRplugin, cons_rplugin, ActiveRpluginMeta
from chromatin.rplugin import venv_package_installed


class Env(Dat['Env']):

    @staticmethod
    def cons(
            rplugins: List[Rplugin]=Nil,
            chromatin_rplugin: Rplugin=None,
            chromatin_venv: VenvMeta=None,
            venvs: Map[str, VenvMeta]=Map(),
            ready: List[str]=Nil,
            active: List[ActiveRpluginMeta]=Nil,
            uninitialized: List[ActiveRpluginMeta]=Nil,
            triggers: Map[str, List[ActiveRpcTrigger]]=Map(),
    ) -> 'Env':
        return Env(
            rplugins,
            Maybe.optional(chromatin_rplugin),
            Maybe.optional(chromatin_venv),
            venvs,
            ready,
            active,
            uninitialized,
            triggers,
        )

    def __init__(
            self,
            rplugins: List[Rplugin],
            chromatin_rplugin: Maybe[Rplugin],
            chromatin_venv: Maybe[VenvMeta],
            venvs: Map[str, VenvMeta],
            ready: List[str],
            active: List[ActiveRpluginMeta],
            uninitialized: List[ActiveRpluginMeta],
            triggers: Map[str, List[ActiveRpcTrigger]],
    ) -> None:
        self.rplugins = rplugins
        self.chromatin_rplugin = chromatin_rplugin
        self.chromatin_venv = chromatin_venv
        self.venvs = venvs
        self.ready = ready
        self.active = active
        self.uninitialized = uninitialized
        self.triggers = triggers

    def add_plugin(self, name: str, spec: str) -> 'Env':
        return self.append1.rplugins(cons_rplugin(name, spec))

    def add_plugins(self, rplugins: List[Rplugin]) -> 'Env':
        return self.append.rplugins(rplugins)

    def add_venv(self, venv: VenvMeta) -> 'Env':
        return self if venv.rplugin in self.venvs else self.append.venvs((venv.rplugin, venv))

    def add_installed(self, rplugin: str) -> 'Env':
        return self.copy(ready=self.ready.cat(rplugin).distinct)

    @property
    def rplugins_with_crm(self) -> List[Rplugin]:
        return self.rplugins.cons_m(self.chromatin_rplugin)

    def rplugin_by_name(self, name: str) -> Either[str, Rplugin]:
        return self.rplugins_with_crm.find(lambda a: a.name == name).to_either(f'no rplugin with name `{name}`')

    @property
    @do(NvimIO[List[Venv]])
    def missing(self) -> Do:
        venvs = yield self.venvs.v.traverse(self.venv_from_meta, NvimIO)
        yield N.pure(venvs.filter_not(venv_package_installed))

    @property
    def _str_extra(self) -> List[Any]:
        return List(self.rplugins, self.venvs)

    @do(NvimIO[Venv])
    def venv_from_meta(self, meta: VenvMeta) -> Do:
        rplugin = yield N.from_either(self.rplugin_by_name(meta.name))
        yield self.venv_dir / L(cons_venv_under)(_, rplugin)

    @property
    def autoreboot(self) -> NvimIO[Boolean]:
        return self.settings.autoreboot.value_or_default

    def rplugin(self, name: str) -> Either[str, Rplugin]:
        return self.rplugins.find(_.name == name).to_either(f'no rplugin named `{name}`')

    @property
    def active_packages(self) -> Either[str, List[Rplugin]]:
        return self.active_rplugins / __.map(_.rplugin)

    @property
    def active_package_names(self) -> Either[str, List[str]]:
        return self.active_rplugins / __.map(_.name)

    @property
    def ready_rplugins(self) -> Either[str, List[Rplugin]]:
        return self.ready.traverse(self.rplugin, Either)

    @property
    def inactive(self) -> List[Rplugin]:
        return self.ready.remove_all(self.active_package_names).traverse(self.rplugin, Either)

    def host_started(self, rplugin: ActiveRpluginMeta) -> 'Env':
        return self.append1.active(rplugin).append1.uninitialized(rplugin)

    def initialization_complete(self) -> 'Env':
        return self.set.uninitialized(List())

    def deactivate_rplugin(self, meta: ActiveRpluginMeta) -> 'Env':
        return self.mod.active(__.without(meta))

    def ready_by_name(self, names: List[str]) -> Either[str, List[Rplugin]]:
        return names.filter(self.ready.contains).traverse(self.rplugin, Either)

    def active_by_name(self, names: List[str]) -> List[ActiveRpluginMeta]:
        return self.active.filter(lambda v: v.rplugin in names)

    def active_rplugin(self, meta: ActiveRpluginMeta) -> Either[str, ActiveRplugin]:
        return (
            self.rplugins
            .find(_.name == meta.rplugin)
            .map(lambda p: ActiveRplugin(p, meta))
            .to_either(f'no rplugin for {meta}')
        )

    def as_active_rplugins(self, meta: List[ActiveRpluginMeta]) -> Either[str, List[ActiveRplugin]]:
        return meta.traverse(self.active_rplugin, Either)

    @property
    def active_rplugins(self) -> Either[str, List[ActiveRplugin]]:
        return self.as_active_rplugins(self.active)

    def active_rplugins_by_name(self, names: List[str]) -> Either[str, List[ActiveRplugin]]:
        return self.as_active_rplugins(self.active_by_name(names))

    @property
    def uninitialized_rplugins(self) -> Either[str, List[ActiveRplugin]]:
        return self.as_active_rplugins(self.uninitialized)

    def add_handlers(self, venv: Venv, triggers: List[ActiveRpcTrigger]) -> 'Env':
        return self.append.triggers((venv.name, triggers))

    def handlers_for(self, plugin: str) -> Either[str, List[ActiveRpcTrigger]]:
        return self.triggers.lift(plugin).to_either(f'no triggers defined for {plugin}')

    @property
    def installed_with_crm(self) -> List[VenvMeta]:
        return self.ready.flat_map(self.venvs.lift).cons_m(self.chromatin_venv)


__all__ = ('Env',)



import abc
import shutil
import venv  # type: ignore
import sys
import pkg_resources
from types import SimpleNamespace

from amino import Path, IO, do, Boolean, env, Either, Try
from amino.boolean import true, false
from amino.do import Do
from amino.dat import ADT, Dat
from amino.logging import module_log

from ribosome.process import Subprocess
from ribosome.logging import ribo_log

from chromatin.model.rplugin import Rplugin, VenvRplugin

log = module_log()
py_exe = f'python{sys.version_info.major}.{sys.version_info.minor}'


class VenvMeta(Dat['VenvMeta']):

    @staticmethod
    def from_ns(rplugin: str, dir: Path, context: SimpleNamespace) -> 'Venv':
        exe = Try(lambda: context.env_exe) / Path
        bin_path = Try(lambda: context.bin_path) / Path
        return VenvMeta(rplugin, dir, exe, bin_path)

    def __init__(
            self,
            rplugin: str,
            dir: Path,
            python_executable: Either[str, Path],
            bin_path: Either[str, Path],
    ) -> None:
        self.rplugin = rplugin
        self.dir = dir
        self.python_executable = python_executable
        self.bin_path = bin_path

    @property
    def name(self) -> str:
        return self.rplugin


class Venv(Dat['Venv']):

    def __init__(self, rplugin: VenvRplugin, meta: VenvMeta) -> None:
        self.rplugin = rplugin
        self.meta = meta

    @property
    def dir(self) -> Path:
        return self.meta.dir

    @property
    def site(self) -> Path:
        return self.dir / 'lib' / py_exe / 'site-packages'

    @property
    def plugin_path(self) -> Path:
        return self.site / self.name / '__init__.py'

    @property
    def name(self) -> str:
        return self.rplugin.name

    @property
    def req(self) -> str:
        return self.rplugin.spec


# TODO remove
class ActiveVenv(Dat['ActiveVenv']):

    def __init__(self, venv: Venv, channel: int, pid: int) -> None:
        self.venv = venv
        self.channel = channel
        self.pid = pid

    @property
    def name(self) -> str:
        return self.venv.name


class VenvStatus(ADT['VenvStatus']):

    @abc.abstractproperty
    def exists(self) -> Boolean:
        ...


class VenvExistent(VenvStatus):

    def __init__(self, plugin: Rplugin, venv: Venv) -> None:
        self.plugin = plugin
        self.venv = venv

    @property
    def exists(self) -> Boolean:
        return true


class VenvAbsent(VenvStatus):

    def __init__(self, plugin: Rplugin) -> None:
        self.plugin = plugin

    @property
    def exists(self) -> Boolean:
        return false


@do(IO[Venv])
def build(global_interpreter: Path, dir: Path, rplugin: VenvRplugin) -> Do:
    interpreter = rplugin.interpreter.get_or_strict(global_interpreter)
    retval, out, err = yield Subprocess.popen(interpreter, '-m', 'venv', str(dir), '--upgrade', timeout=30)
    success = retval == 0
    yield (
        IO.delay(cons_venv, dir, rplugin)
        if success else
        IO.failed(f'creating venv for {rplugin}: {err.join_lines}')
    )


@do(IO[Venv])
def cons_venv(dir: Path, rplugin: VenvRplugin) -> Do:
    builder = yield IO.delay(venv.EnvBuilder, system_site_packages=False, with_pip=True)
    context = yield IO.delay(builder.ensure_directories, str(dir))
    return Venv(rplugin, VenvMeta.from_ns(rplugin.name, dir, context))


def cons_venv_under(base_dir: Path, rplugin: VenvRplugin) -> IO[Venv]:
    return cons_venv(base_dir / rplugin.name, rplugin)


@do(IO[None])
def remove_dir(dir: Path) -> Do:
    exists = yield IO.delay(dir.exists)
    yield IO.delay(shutil.rmtree, dir) if exists else IO.pure(None)


@do(IO[Venv])
def bootstrap(interpreter: Path, base_dir: Path, rplugin: VenvRplugin) -> Do:
    venv_dir = base_dir / rplugin.name
    log.debug(f'bootstrapping {rplugin} in {venv_dir}')
    yield remove_dir(venv_dir)
    yield build(interpreter, venv_dir, rplugin)


class VenvPackageStatus(ADT['VenvPackageStatus']):

    @abc.abstractproperty
    def exists(self) -> Boolean:
        ...


class VenvPackageExistent(VenvPackageStatus):

    def __init__(self, venv: Venv, dist: pkg_resources.Distribution) -> None:
        self.venv = venv
        self.dist = dist

    @property
    def exists(self) -> Boolean:
        return true


class VenvPackageAbsent(VenvPackageStatus):

    def __init__(self, venv: Venv) -> None:
        self.venv = venv

    @property
    def exists(self) -> Boolean:
        return false

__all__ = ('VenvStatus', 'VenvExistent', 'VenvAbsent', 'VenvPackageAbsent', 'VenvPackageExistent', 'VenvPackageStatus',
           'Venv', 'ActiveVenv', 'cons_venv_under')
